/**
 * The java media client class
 * InternetPhone.java
 * Dec 14, 2001
 */
import java.io.*;
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.net.*;

import javax.media.*;
import javax.media.rtp.*;
import javax.media.rtp.event.*;
import javax.media.rtp.rtcp.*;
import javax.media.protocol.*;
import javax.media.protocol.DataSource;
import javax.media.format.AudioFormat;
import javax.media.Format;
import javax.media.format.FormatChangeEvent;
import javax.media.control.*;

/**
 * InternetPhone using JMF 2.1.1 build a simple "Internet telephone" 
 * that can send audio, using RTP, to a given IP address or host name and 
 * receive audio. The audio formats, different packetization intervals 
 * audio input and output source are selectable from the command line.
 */
public class InternetPhone
    implements ReceiveStreamListener, ControllerListener {

    private int localPort;          
    private String remoteAddress;
    private int remotePort;
    
    private String codec;
    private double sampleRate;
    private int sampleSizeInBits;
    private int channel;
    private int packetizationPeriod;

    private String inputFile;
    private String outputFile;

    // The processor and datasink are used to send audio data to network
    DataSink datasink = null;
    Processor processor = null;

    /**
     * mgr, audioPlayer, filewriter and sh are used to receive 
     * data from network
     */
    RTPManager mgr;    
    Player audioPlayer;
    DataSink filewriter = null;
    StateHelper sh = null;

    public InternetPhone(String remoteAddress, int localPort, int remotePort){

        this.localPort = localPort;
        this.remoteAddress = remoteAddress;
        this.remotePort = remotePort;
        codec = "LINEAR";                     // default codec = LINEAR
        sampleRate = 44100;                    // default sampleRate =  8000
        sampleSizeInBits = 16;                // default sampleSizeInBits = 16
        channel = 2;                          // default channel = Stereo
        packetizationPeriod = 60;             // default value  = 60ms

    }

    // set the codec
    public void setCodec(String codec) {
        this.codec = codec;
    }

    // set the sampleRate
    public void setSampleRate(double sampleRate) {
        this.sampleRate = sampleRate;
    }

    // set packetizationPeriod
    //public void setPacketizationPeriod(int packetizationPeriod) {
    //    this.packetizationPeriod = packetizationPeriod;
    //}
    
    // set inputFile
    public void setInputFile(String filename) {
        this.inputFile = filename;
    }
    
    //set outputFile
    public void setOutputFile(String filename) {
        this.outputFile = filename;
    }

    // Start the transmit
    public synchronized String start() {
        String errorMessage;

        // Create a processor for the specified media locator
        // and program it to output
        errorMessage = createProcessor();
        if (errorMessage != null) {
	    destroyProcessor();
	    return errorMessage;
	}

        // Create an RTP session to transmit the output of the
        // processor to the specified IP address and port no.
        
        // Start the transmission
	System.out.println("Starting the processor");
	processor.start();

        return null;
    }

    // destroy the processor
    private void destroyProcessor() {
	if (processor == null)
	    return;
	processor.close();
	processor = null;
    }


    /**
     * Implement the SessionListener's method "update".
     */
    /*
    public synchronized void update(SessionEvent evt) {
	if (evt instanceof NewParticipantEvent) {
	    Participant p = ((NewParticipantEvent)evt).getParticipant();
	    System.err.println("  -A new participant had just joined: " + 
                               p.getCNAME());
	}
    }
    */

    /**
     * Implement the ReceiveStreamListener's method "update"
     */
    public synchronized void update(ReceiveStreamEvent evt) {

	RTPManager mgr = (RTPManager)evt.getSource();
	Participant participant = evt.getParticipant();	// could be null.
	ReceiveStream stream = evt.getReceiveStream();  // could be null.
        Player p;

	if (evt instanceof RemotePayloadChangeEvent) {
     
	    System.err.println("  -Received an RTP PayloadChangeEvent.");
	    System.err.println("    Cannot handle payload change.");
	    System.exit(0);

	}
    
	else if (evt instanceof NewReceiveStreamEvent) {

	    try {
		stream = ((NewReceiveStreamEvent)evt).getReceiveStream();
		DataSource ds = stream.getDataSource();

		// Find out the formats.
		RTPControl ctl = 
                    (RTPControl)ds.getControl("RTPControl");
		if (ctl != null)
		    System.err.println("  -Recevied new RTP stream: " + 
                                       ctl.getFormat());
		else
		    System.err.println("  -Recevied new RTP stream");


		if (participant == null)
		    System.err.println("      The sender of this stream " + 
                                       "had yet to be identified.");
		else {
		    System.err.println("      The stream comes from: " + 
                                       participant.getCNAME()); 
		}

                if (outputFile == null) {

                    // no output option create a player to 
                    // player the received audio data
                    audioPlayer = Manager.createPlayer(ds);

                    if (audioPlayer == null)
                        return;
                    
                    audioPlayer.addControllerListener(this);
                    audioPlayer.realize();

                } 
                else {
                    // the output option is present
                    // writing received data to outputFile
                    FileTypeDescriptor ftd = 
                        new FileTypeDescriptor(FileTypeDescriptor.BASIC_AUDIO);

                    try {
                        // handling this ds over to a file datasink
                        String output = 
                            "file:/" + new File(outputFile).getAbsolutePath();
            
                        MediaLocator outlocator = new MediaLocator(output);
                        Processor p1 = Manager.createProcessor(ds);

                        sh = new StateHelper(p1);
                        sh.configure(100000);

                        //System.out.println("setContentDescriptor");
                        p1.setContentDescriptor(ftd);
                        //System.out.println("sh realize");
                        sh.realize(10000);
                        
                        DataSource s = p1.getDataOutput();
                        
                        //filewriter = null;
                        //create datasink success
                        filewriter = Manager.createDataSink(s, outlocator);
	    		
                        filewriter.open();
                
                        StreamWriterControl swc = (StreamWriterControl)
                            p1.getControl("StreamWriterControl");

			//set limit to 5MB
                        if (swc != null)
                            swc.setStreamSizeLimit(5*1000*1000);
			
			// now start the filewriter
                        try {
                            filewriter.start();
                        } catch (IOException e) {
                            System.exit(-1);
                        }

 			//Wait for an EndOfStream from the DataSink and close
                        sh.playToEndOfMedia();

                    } 
                    catch (IOException ioe) {
                        System.err.println("IOException in update" +
                                           ioe.getMessage());
                        return;
                    } 
                    catch (NoDataSinkException ndse) {
                        System.err.println("NoDataSinkException in update" + 
                                           ndse.getMessage());
                        return;
                    } 
                    catch (SecurityException se) {
                        System.err.println("SecurityException in update" + 
                                           se.getMessage());
                        return;
                    } 
                    catch (Exception e) {
                        System.err.println("Exception in update:" + 
                                           e.getMessage());
                        return;
                    }
                }   
            } 
            catch (Exception e) {
                System.err.println("newReceiveStreamEvent exception " 
                                   + e.getMessage());
                return;
            }
	}
        
	else if (evt instanceof StreamMappedEvent) {
            
            if (stream != null && stream.getDataSource() != null) {
		DataSource ds = stream.getDataSource();

		// Find out the formats.
		RTPControl ctl = 
                    (RTPControl)ds.getControl("javax.media.rtp.RTPControl");

		System.out.println("  -The previously unidentified stream ");

		if (ctl != null) {
		    System.out.println("     " + ctl.getFormat());
                    System.out.println("     had now been identified " + 
                                       "as sent by: "+ participant.getCNAME());
                }
            }
	}

	else if (evt instanceof ByeEvent) {
            
            System.out.println("  -Got \"bye\" from: " + 
                               participant.getCNAME());

            // close the audioPlayer ot the filewriter         
            if (audioPlayer != null)
		audioPlayer.close();

            if(sh != null)
                sh.close();

            if(filewriter != null)
                filewriter.close();
	}
        
    }


    /**
     * ControllerListener for the Players.
     */
    public synchronized void controllerUpdate(ControllerEvent evt) {

        Player player = null;
        Controller controller = (Controller)evt.getSource();

        if (controller instanceof Player)
            player  = (Player)evt.getSource();

        if (player == null)
            return;

	if (evt instanceof RealizeCompleteEvent) {
	    System.out.println("Player enters the Realized state");
            player.start();
	}
	if (evt instanceof PrefetchCompleteEvent) {
	    System.out.println("Player enters the Prefetched state");
	}
	if (evt instanceof FormatChangeEvent) {
	    System.out.println("Format change occurred");
	}

	if (evt instanceof ControllerErrorEvent) {

	    player.removeControllerListener(this);

	    System.err.println("Receive internal error: " + evt);
	}

    }


    /**
     * Stops the transmission if already started
     */
    public void stop() {
        if((datasink == null)||(processor == null)){
            System.err.println("You must start transmitter before stop!");
            return;
        }
        synchronized (this) {
            // close the process
            processor.stop();
            processor.close();
            processor = null;
            datasink.close();
            datasink = null;

            // close RTPManager
            closeRTPManager();

        }
    }


    /**
     * create the processor.
     */
    private String createProcessor() {
        AudioFormat audioFormat = 
            new AudioFormat(AudioFormat.ULAW_RTP, 8000, 8, 1);
        
        TrackControl tracks[];
        Control controls[];
        boolean found = false;
        DataSource datasource;
        datasink = null;

        // Set the output content descriptor to RAW_RTP
        ContentDescriptor cd = 
            new ContentDescriptor(ContentDescriptor.RAW_RTP);

        if (inputFile == null) {
            // inputFile option is not present.
            // Get data from an audio device
            MediaLocator capturedDeciceLacator = getMediaLocator();

            if(capturedDeciceLacator == null)
                return "Couldn't find the relevant audio capture device";

            System.out.println("Creating the processor...");

            try {
                processor = 
                    Manager.createProcessor(capturedDeciceLacator); 
            }
            catch (NoProcessorException npe) {
                System.out.println("No processor Exception: " 
                                   + npe.getMessage());
                return "Failed to create Processor";
            }
            catch (IOException ioe) {
                System.out.println("IOException: " + ioe.getMessage());
                return "IOException while creating Processor";
            }
            
            if(waitForState(processor, Processor.Configured) == false)
	    	return "Couldn't configure processor";
            
            tracks = processor.getTrackControls();
            
            if (tracks == null || tracks.length < 1) {
                return "Coundn't find tracks in processor";
            }

            // Set one one track to ULAW RTP format. Disable
            // all other tracks.
            for (int i = 0; i < tracks.length; i++) { 
                if (!found && tracks[i] instanceof FormatControl) {  
                    if(((FormatControl)tracks[i]).setFormat(audioFormat)==null)
                       tracks[i].setEnabled(false);
                    else{
                        found = true;
                    }
                } 
                else { 
                    tracks[i].setEnabled(false); 
                } 
            }

            if(!found)
                return "Could not set format in any track in processor";

            processor.setContentDescriptor(cd);

            if(waitForState(processor, Processor.Realized) == false)
                return "Couldn't realize processor"; 
            
            try { 
                datasource = processor.getDataOutput(); 
            } 
            catch (Exception e) {
               System.out.println("Exception: " + e.getMessage());
                e.printStackTrace();
                return "Couldn't retrieve the processor output";
            }
            
            try {
                String url= "rtp://"+remoteAddress+":" + remotePort 
                    + "/audio/";
                
                datasink = 
                    Manager.createDataSink(datasource, new MediaLocator(url));
                datasink.open(); 
                datasink.start();
                datasource.start();
            } 
            catch (NoDataSinkException ndse) {
                System.out.println("NoDataSinkException: " 
                                   + ndse.getMessage());
                return "NoDataSinkException in creatProcessor";
            } 
            catch (IOException ioe) {
                System.out.println("IOException: " + ioe.getMessage());
                return "IOException in creatProcessor";
            }
        }
        else {
            // inputfile option is present
            // Get data from inputFile
            String input = "file:/" + new File(inputFile).getAbsolutePath();

            MediaLocator locator = new MediaLocator(input);
            DataSource ds;
            
            try {
                ds = Manager.createDataSource(locator);
            } 
            catch (NoDataSourceException ndse) {
                System.out.println("NoDataSinkException: " 
                                   + ndse.getMessage());
                return "NoDataSourceException in createProcessor";
            } 
            catch (IOException ioe) {
                System.out.println("IOException: " + ioe.getMessage());
                return "IOException in createProcessor when create datasource";
            }

            System.out.println("Creating the processor to for the inputfile");
            try {
                processor = Manager.createProcessor(ds);
            } 
            catch (NoProcessorException npe) {
                System.out.println("NoProcessorException: " 
                                   + npe.getMessage());
                return "Failed to create Processor for the input file";
            } 
            catch (IOException ioe) {
                System.out.println("IOException: " + ioe.getMessage());
                return "Failed to create Processor for the inputfile";
            }

            // Wait for it to configure
            if (waitForState(processor, Processor.Configured) == false)
                return "Couldn't configure processor";

            processor.setContentDescriptor(cd);

            // Get the tracks from the processor
            tracks = processor.getTrackControls();

            if (tracks == null || tracks.length < 1)
                return "Couldn't find tracks in processor";

            Format supported[];
            Format chosen;
            boolean atLeastOneTrack = false;

            // Program the tracks.
            for (int i = 0; i < tracks.length; i++) {
                Format format = tracks[i].getFormat();
                if (tracks[i].isEnabled()) {
                    
                    supported = tracks[i].getSupportedFormats();
                    
                    if (supported.length > 0) {
                        chosen = supported[0];
                        tracks[i].setFormat(chosen);
                        System.out.println("Track " + i + " is set to " + 
                                           "transmit as:\n   " + chosen);
                        atLeastOneTrack = true;
                    } else
                        tracks[i].setEnabled(false);
                } else
                    tracks[i].setEnabled(false);
            }
            
            if (!atLeastOneTrack)
                return "Couldn't set any of the tracks to a valid RTP format";

            // Wait for it to realize
            if (waitForState(processor, Controller.Realized) == false)
                return "Couldn't realize processor";

            try { 
                datasource = processor.getDataOutput(); 
            } 
            catch (Exception e) {
                System.out.println("Exception: " + e.getMessage());
                return "Couldn't retrieve the processor output";
            }
            
            try {
                String url= "rtp://"+remoteAddress+":"+remotePort+"/audio/";
                datasink = 
                    Manager.createDataSink(datasource, new MediaLocator(url));
                datasink.open();
                datasink.start();
                datasource.start();
            } 
            catch (NoDataSinkException nde) {
                System.out.println("NoDataSinkException: " + nde.getMessage());
                return "NoDataSinkException in creatProcessor";
            } 
            catch (IOException ioe) {
                System.out.println("IOException: " + ioe.getMessage());
                return "IOException in creatProcessor";
            }
        }
        return null;
    }

    /**
     * Get MediaLocator for audiodevice. If no audiodevice exists return null.
     */
    private MediaLocator getMediaLocator() {

        AudioFormat audioFormat = 
            new AudioFormat(codec, sampleRate, sampleSizeInBits, channel);

        // Get all audio device which support this format 
        Vector devices = CaptureDeviceManager.getDeviceList(audioFormat);
        
        if(devices.size() == 0)
            return null;
        
        return ((CaptureDeviceInfo) devices.elementAt(0)).getLocator();
        
    }

    /**
     * create RTPManager
     */
    public String createRTPManager() {
        try {
	    InetAddress remoteIPAddr = InetAddress.getByName(remoteAddress);
	    SessionAddress localAddr;
	    SessionAddress remoteAddr;
	    
	    // create RTPManger
            mgr = (RTPManager) RTPManager.newInstance();
            mgr.addReceiveStreamListener(this);
            
            // generate the local session address and remote session address
            localAddr= new SessionAddress(InetAddress.getLocalHost(),
                                          localPort);
            remoteAddr = new SessionAddress(remoteIPAddr, remotePort);

            // initialize the RTPManger
            mgr.initialize(localAddr);
            // add target to the RTPManager
            mgr.addTarget(remoteAddr);

        } 
        catch (InvalidSessionAddressException e) {
            System.out.println("InvalidSessionAddressException: " 
                               + e.getMessage());
            return "InvalidSessionAddressException in createRTPManager";
        } 
        catch (IOException ioe){
            System.out.println("IOException: " + ioe.getMessage());
            return "IOException in createRTPManager";
        }

        return null;
    }


    /**
     * close the RTPManager
     */
    public void closeRTPManager() {
        if (audioPlayer != null)
            audioPlayer.close();
        
        if(sh != null)
            sh.close();
        
        if (filewriter != null)
            filewriter.close();

        mgr.removeTargets( "Closing RTPManager");
        mgr.dispose();
        mgr = null;
    }
    

    /****************************************************************
     * Convenience methods to handle processor's state changes.
     ****************************************************************/
    
    private Integer stateLock = new Integer(0);
    private boolean failed = false;
    
    Integer getStateLock() {
	return stateLock;
    }
    
    void setFailed() {
	failed = true;
    }
    
    private synchronized boolean waitForState(Processor p, int state) {
	p.addControllerListener(new StateListener());
	failed = false;
        
	// Call the required method on the processor
	if (state == Processor.Configured) {
	    p.configure();
	} else if (state == Processor.Realized) {
	    p.realize();
	}
	
	// Wait until we get an event that confirms the
	// success of the method, or a failure event.
	// See StateListener inner class
	while (p.getState() < state && !failed) {
	    synchronized (getStateLock()) {
		try {
		    getStateLock().wait();
		} catch (InterruptedException ie) {
		    return false;
		}
	    }
	}

	if (failed)
	    return false;
	else
	    return true;
    }
    
    /****************************************************************
     * Inner Classes
     ****************************************************************/

    class StateListener implements ControllerListener {

	public void controllerUpdate(ControllerEvent ce) {
            
	    // If there was an error during configure or
	    // realize, the processor will be closed
	    if (ce instanceof ControllerClosedEvent)
		setFailed();
            
	    // All controller events, send a notification
	    // to the waiting thread in waitForState method.
	    if (ce instanceof ControllerEvent) {
		synchronized (getStateLock()) {
		    getStateLock().notifyAll();
		}
	    }
	}
    }
}

/**
 * StateHelper is a helper class that implements the ControllerListener 
 * interface and can be used to manage the state of a Processor. 
 */
class StateHelper implements ControllerListener {

    Player player = null;
    boolean configured = false;
    boolean realized = false;
    boolean prefetched = false;
    boolean eom = false;
    boolean failed = false;
    boolean closed = false;
    
    public StateHelper(Player p) {
	player = p;
	p.addControllerListener(this);
    }

    public boolean configure(int timeOutMillis) {
	long startTime = System.currentTimeMillis();
	synchronized (this) {
	    if (player instanceof Processor)
		((Processor)player).configure();
            else
		return false;

	    while (!configured && !failed) {
		try {
		    wait(timeOutMillis);
		} catch (InterruptedException ie) {
		}
		if (System.currentTimeMillis() - startTime > timeOutMillis)
		    break;
	    }
	}
	return configured;
    }
    
    public boolean realize(int timeOutMillis) {
	long startTime = System.currentTimeMillis();
	synchronized (this) {
	    player.realize();
	    while (!realized && !failed) {
		try {
		    wait(timeOutMillis);
		} catch (InterruptedException ie) {
		}
		if (System.currentTimeMillis() - startTime > timeOutMillis)
		    break;
	    }
	}
	return realized;
    }

    public boolean prefetch(int timeOutMillis) {
	long startTime = System.currentTimeMillis();
	synchronized (this) {
	    player.prefetch();
	    while (!prefetched && !failed) {
		try {
		    wait(timeOutMillis);
		} catch (InterruptedException ie) {
		}
		if (System.currentTimeMillis() - startTime > timeOutMillis)
		    break;
	    }
	}
	return prefetched && !failed;
    }

    public boolean playToEndOfMedia() {

	synchronized (this) {
	    player.start();
		try {
		    wait();
		} catch (InterruptedException ie) {
		}
	}

	return true;
    }

    public void close() {
	synchronized (this) {
	    player.close();
	    while (!closed) {
		try {
		    wait(100);
		} catch (InterruptedException ie) {
		}
                notifyAll();
	    }
	}
	player.removeControllerListener(this);
    }

    public synchronized void controllerUpdate(ControllerEvent ce) {
	if (ce instanceof RealizeCompleteEvent) {
	    realized = true;
	} else if (ce instanceof ConfigureCompleteEvent) {
	    configured = true;
	} else if (ce instanceof PrefetchCompleteEvent) {
	    prefetched = true;
	} else if (ce instanceof EndOfMediaEvent) {
	    eom = true;
	} else if (ce instanceof ControllerErrorEvent) {
	    failed = true;
	} else if (ce instanceof ControllerClosedEvent) {
	    closed = true;
	} else {
	    return;
	}
	notifyAll();
    }
}



