/**
 * The java media client class
 * JMClient.java
 * Dec 14, 2001
 */
import java.net.*;

public class JMClient{

    InternetPhone phone;

    public JMClient(ParseCommandLine options){

        phone = 
            new InternetPhone(options.getRemoteAddress(),
                              options.getLocalPort(),
                              options.getRemotePort());
        
        if (options.IsPresent("c")) {
            /* Option c is present */
            System.out.println("codec:"+options.GetValue("c"));
            String codec = options.GetValue("c");
            if(codec != null) 
                phone.setCodec(codec);
        }

        if (options.IsPresent("s")) {
            /* Option s is present */
            System.out.println("samplerate:"+options.GetValue("s"));
            long sampleRate = options.GetLong("s");
            if(sampleRate > 0)
               phone.setSampleRate(sampleRate);
        }

        //if (options.IsPresent("p")) {
            /* Option p is present */
        //    System.out.println("packetization:"+options.GetValue("p"));
        //    int packetizationPeriod = options.GetInteger("p");
        //    if(packetizationPeriod > 0)
        //        phone.setPacketizationPeriod(packetizationPeriod);
        //}

        if (options.IsPresent("f")) {
            /* Option f is present */
            System.out.println("inputFile: "+options.GetValue("f"));
            String inputFile = options.GetValue("f");
            if(inputFile != null)
               phone.setInputFile(inputFile);
        }

        if (options.IsPresent("o")) {
            /* Option o is present */
            System.out.println("outputFile:"+options.GetValue("o"));
            String outputFile = options.GetValue("o");
            if(outputFile != null)
               phone.setOutputFile(outputFile);
        }

        String errorMessage;
        
        System.out.println("Creating the RTPmanager...");

        if((errorMessage = phone.createRTPManager()) != null) {
            System.err.println(errorMessage);
            System.exit(0);
        }

        if((errorMessage = phone.start()) != null) {
            System.err.println(errorMessage);
            System.exit(0);
        }

    }

    public static void main(String args[]) throws Exception{

        // Parse command line
        ParseCommandLine options = new ParseCommandLine(args);
        
        JMClient client = new JMClient(options);
        
    }
}


/**
 * ParseCommandLine parses the command line of the program and
 * provides functions to extract the various options, whether
 * they are present or not.
 */
class ParseCommandLine {
    String cmdline[];
    String remoteAddress;
    int localPort;
    int remotePort;

    //ParseCommandLine Constructor
    public ParseCommandLine(String args[]){
	cmdline = new String[args.length];

	for ( int i=0; i<args.length; i++){
            cmdline[i] = args[i];
	}

        if(cmdline.length > 0) {
            parseAddressandPort(cmdline[cmdline.length-1]);
        }else{
            printUsage();
        }
    }

    /* Return remote address, local port and remote port */
    public String getRemoteAddress(){
        return remoteAddress;
    }

    public int getLocalPort(){
        return localPort;
    }

    public int getRemotePort(){
        return remotePort;
    }

    /*
     * Check whether an option is present or not.
     * e.g., IsPresent("r") will check is -r is
     * present on command line or not.
     */
    public boolean IsPresent(String option) {
	for (int i=0; i<cmdline.length; i++) {
	    if ( cmdline[i].compareTo("-" + option )==0) {
		return true;
	    }
	}
	return false;
    }

    /*
     * Get the value of the option if it needs a value.
     * for example "-o filename" is a possible option
     * then GetValue("o") will return the filename
     * given on the command line.
     */
    public String GetValue(String option) {
	for (int i = 0; i < cmdline.length - 1; i++) {
	    if (cmdline[i].compareTo("-" + option)==0) {
		return cmdline[i+1];
	    }
	}
	return null;
    }

    /*
     * Get the value of the option as an integer.
     * for example if "-p packetization-period" is a possible
     * option then GetInteger("t") returns the int value
     * of the timeout option. Returns -1 on error and
     * -2 if not present.
     */
    public int GetInteger(String option) {
	for (int i=0; i<cmdline.length - 1; i++) {
	    if ( cmdline[i].compareTo("-" + option )==0) {
		int size;
		try {
		    size = Integer.parseInt(cmdline[i+1]);
		}catch(NumberFormatException e) {
                    System.err.println(e);
                    printUsage();
		    return -1; /* -1 is returned for error */
		}
		return size;
	    }
	}
	return -2; /* If not present */
    }

    /*
     * Get the value of the option as an long.
     * for example if "-s sample-rate" is a possible
     * option then GetLong("t") returns the long value
     * of the timeout option. Returns -1 on error and
     * -2 if not present.
     */
    public long GetLong(String option) {
	for (int i=0; i<cmdline.length - 1; i++) {
	    if ( cmdline[i].compareTo("-" + option )==0) {
		long size;
		try {
		    size = Long.parseLong(cmdline[i+1]);
		}catch(NumberFormatException e) {
                    /* ... catch the exception ... */
		    return -1; /* -1 is returned for error */
		}
		return size;
	    }
	}
	return -2; /* If not present */
    }

    /* Parse the address and port */
    private void parseAddressandPort(String IPAndPorts){
        int index = 0;
        if(IPAndPorts.indexOf("/", index) > 0){
            remoteAddress = 
                IPAndPorts.substring(index,IPAndPorts.indexOf("/"));
            index = IPAndPorts.indexOf("/", index) + 1;

            try{
                InetAddress.getByName(remoteAddress);
            }catch(UnknownHostException e){
                System.err.println(e);
                printUsage();       
                System.exit(-1);
            }

            if(IPAndPorts.indexOf("/", index) > 0){
                //user has specified local-port/remote-port
                String lport = 
                    IPAndPorts.substring(index, 
                                         IPAndPorts.indexOf("/",index));
                String rport = 
                    IPAndPorts.substring(IPAndPorts.indexOf("/",index)+1);
                try{
                    localPort = Integer.parseInt(lport);
                    remotePort = Integer.parseInt(rport);
                }catch(NumberFormatException e) {
                    System.err.println(e);
                    printUsage();
                    System.exit(-1);
                }
            }else{
                //user only specified remote-port set localport = remoteport
                String lport = IPAndPorts.substring(index);
                try{
                    localPort = Integer.parseInt(lport);
                    remotePort = localPort;
                }catch(NumberFormatException e) {
                    System.err.println(e);
                    printUsage();
                    System.exit(-1);
                }
            }
        }else{
            printUsage();
            System.exit(-1);
        }
    }

    /* print the jmclient usage */
    private void printUsage(){
        System.err.println("Usage:\n" +
                           "java jmclient [-c codec] [-s sample-rate]" +
                           "  [-p packetization-period]\n" +
                           "  [-f input-file] [-o output-file]\n" +
                           "  <remote-ip>/[local-port/]<remote-local-port>");
        System.exit(-1);
    }
} //ParseCommandLine


  




